from __future__ import annotations
from typing import Callable, Generic, Iterable, Iterator, TypeVar
from math import frexp, isfinite, isnan, ldexp
import unittest


_V = TypeVar("_V")

class _QuadTreeNode(Generic[_V]):
    """ node in a quad-tree extending from `-(2 ** log2_size)` to
    `2 ** log2_size` exclusive on both sides in both `x` and `y`
    dimensions.

    invariants:
    * if `self.values is not None` then all of
      `nx_ny`, `px_ny`, `nx_py`, and `px_py` are `None`.
    * the `log2_size` field of all of `nx_ny`, `px_ny`, `nx_py`, and `px_py`
      is always `self.log2_size - 1`.
    """
    __slots__ = "log2_size", "nx_ny", "px_ny", "nx_py", "px_py", "values"
    log2_size: int
    nx_ny: None | _QuadTreeNode[_V]
    px_ny: None | _QuadTreeNode[_V]
    nx_py: None | _QuadTreeNode[_V]
    px_py: None | _QuadTreeNode[_V]
    values: None | list[tuple[float, float, _V]]

    def __init__(
        self, *,
        log2_size: int,
        nx_ny: None | _QuadTreeNode[_V],
        px_ny: None | _QuadTreeNode[_V],
        nx_py: None | _QuadTreeNode[_V],
        px_py: None | _QuadTreeNode[_V],
        values: None | list[tuple[float, float, _V]],
    ):
        self.log2_size = log2_size
        self.nx_ny = nx_ny
        self.px_ny = px_ny
        self.nx_py = nx_py
        self.px_py = px_py
        self.values = values

    @staticmethod
    def leaf(
        log2_size: int,
        values: None | list[tuple[float, float, _V]] = None,
    ) -> _QuadTreeNode[_V]:
        if values is None:
            values = []
        return _QuadTreeNode(
            log2_size=log2_size,
            nx_ny=None,
            px_ny=None,
            nx_py=None,
            px_py=None,
            values=values,
        )

    @staticmethod
    def interior(
        log2_size: int, *,
        nx_ny: None | _QuadTreeNode[_V] = None,
        px_ny: None | _QuadTreeNode[_V] = None,
        nx_py: None | _QuadTreeNode[_V] = None,
        px_py: None | _QuadTreeNode[_V] = None,
    ) -> _QuadTreeNode[_V]:
        return _QuadTreeNode(
            log2_size=log2_size,
            nx_ny=nx_ny,
            px_ny=px_ny,
            nx_py=nx_py,
            px_py=px_py,
            values=None,
        )

    def __iter__(self) -> Iterator[tuple[float, float, _V]]:
        if self.nx_ny is not None:
            yield from self.nx_ny
        if self.px_ny is not None:
            yield from self.px_ny
        if self.nx_py is not None:
            yield from self.nx_py
        if self.px_py is not None:
            yield from self.px_py
        if self.values is not None:
            yield from self.values

    def __repr__(self, indent=0) -> str:
        nl_indent = '\n' + ' ' * (4 * indent)
        parts = [f"    log2_size={self.log2_size}"]
        if self.nx_ny is not None:
            parts.append(f"    nx_ny={self.nx_ny.__repr__(indent + 1)}")
        if self.px_ny is not None:
            parts.append(f"    px_ny={self.px_ny.__repr__(indent + 1)}")
        if self.nx_py is not None:
            parts.append(f"    nx_py={self.nx_py.__repr__(indent + 1)}")
        if self.px_py is not None:
            parts.append(f"    px_py={self.px_py.__repr__(indent + 1)}")
        fn = "interior"
        if self.values is not None:
            if len(self.values) == 0:
                parts.append("    values=[]")
            else:
                prefix = "    values=[" + nl_indent
                for i in self.values:
                    parts.append(f"{prefix}        {i}")
                    prefix = ""
                parts.append("    ]")
            fn = "leaf"
        parts.append(")")
        sep = ',' + nl_indent
        return f"{fn}({nl_indent}{sep.join(parts)}"

    def split(
        self,
        split_x: float,
        split_y: float,
        delta: float,
        max_leaf_size: int, *,
        trace: None | Callable[[str]],
    ):
        assert self.values is not None, "can't split interior node"
        if len(self.values) <= max_leaf_size:
            return
        values = self.values
        self.values = None  # convert to interior node
        for i in values:
            x, y, _ = i
            if x < split_x:
                if y < split_y:
                    if self.nx_ny is None:
                        self.nx_ny = _QuadTreeNode.leaf(log2_size=self.log2_size - 1)
                    assert self.nx_ny.values is not None
                    self.nx_ny.values.append(i)
                else:
                    if self.nx_py is None:
                        self.nx_py = _QuadTreeNode.leaf(log2_size=self.log2_size - 1)
                    assert self.nx_py.values is not None
                    self.nx_py.values.append(i)
            else:
                if y < split_y:
                    if self.px_ny is None:
                        self.px_ny = _QuadTreeNode.leaf(log2_size=self.log2_size - 1)
                    assert self.px_ny.values is not None
                    self.px_ny.values.append(i)
                else:
                    if self.px_py is None:
                        self.px_py = _QuadTreeNode.leaf(log2_size=self.log2_size - 1)
                    assert self.px_py.values is not None
                    self.px_py.values.append(i)
        if trace is not None:
            trace(f"split leaf split_x={split_x} "
                f"split_y={split_y} delta={delta} self={self}")
        if self.nx_ny is not None:
            self.nx_ny.split(
                split_x=split_x - delta,
                split_y=split_y - delta,
                delta=delta * 0.5,
                max_leaf_size=max_leaf_size,
                trace=trace,
            )
        if self.nx_py is not None:
            self.nx_py.split(
                split_x=split_x - delta,
                split_y=split_y + delta,
                delta=delta * 0.5,
                max_leaf_size=max_leaf_size,
                trace=trace,
            )
        if self.px_ny is not None:
            self.px_ny.split(
                split_x=split_x + delta,
                split_y=split_y - delta,
                delta=delta * 0.5,
                max_leaf_size=max_leaf_size,
                trace=trace,
            )
        if self.px_py is not None:
            self.px_py.split(
                split_x=split_x + delta,
                split_y=split_y + delta,
                delta=delta * 0.5,
                max_leaf_size=max_leaf_size,
                trace=trace,
            )

    def range(
        self, *,
        min_x: float,
        max_x: float,
        min_y: float,
        max_y: float,
        node_min_x: float,
        node_max_x: float,
        node_min_y: float,
        node_max_y: float,
        trace: None | Callable[[str]],
    ) -> Iterable[tuple[float, float, _V]]:
        if self.values is not None:
            # leaf node
            if trace is not None:
                trace(
                    f"range(\n"
                    f"    min_x={min_x}, max_x={max_x},\n"
                    f"    min_y={min_y}, max_y={max_y},\n"
                    f"    node_min_x={node_min_x}, node_max_x={node_max_x},\n"
                    f"    node_min_y={node_min_y}, node_max_y={node_max_y},\n"
                    f"): leaf {self}"
                )
            for i in self.values:
                x, y, _ = i
                if min_x <= x <= max_x and min_y <= y <= max_y:
                    if trace is not None:
                        trace(f"yielding {i}")
                    yield i
                else:
                    if trace is not None:
                        trace(f"skipping {i}")
            return
        # interior node
        split_x = (node_min_x + node_max_x) * 0.5
        split_y = (node_min_y + node_max_y) * 0.5
        if trace is not None:
            trace(
                f"range(\n"
                f"    min_x={min_x}, max_x={max_x},\n"
                f"    min_y={min_y}, max_y={max_y},\n"
                f"    node_min_x={node_min_x}, node_max_x={node_max_x},\n"
                f"    node_min_y={node_min_y}, node_max_y={node_max_y},\n"
                f"): interior split_x={split_x} split_y={split_y} {self}"
            )
        if self.nx_ny is not None and min_x <= split_x and min_y <= split_y:
            yield from self.nx_ny.range(
                min_x=min_x,
                max_x=max_x,
                min_y=min_y,
                max_y=max_y,
                node_min_x=node_min_x,
                node_max_x=split_x,
                node_min_y=node_min_y,
                node_max_y=split_y,
                trace=trace,
            )
        if self.nx_py is not None and min_x <= split_x and max_y >= split_y:
            yield from self.nx_py.range(
                min_x=min_x,
                max_x=max_x,
                min_y=min_y,
                max_y=max_y,
                node_min_x=node_min_x,
                node_max_x=split_x,
                node_min_y=split_y,
                node_max_y=node_max_y,
                trace=trace,
            )
        if self.px_ny is not None and max_x >= split_x and min_y <= split_y:
            yield from self.px_ny.range(
                min_x=min_x,
                max_x=max_x,
                min_y=min_y,
                max_y=max_y,
                node_min_x=split_x,
                node_max_x=node_max_x,
                node_min_y=node_min_y,
                node_max_y=split_y,
                trace=trace,
            )
        if self.px_py is not None and max_x >= split_x and max_y >= split_y:
            yield from self.px_py.range(
                min_x=min_x,
                max_x=max_x,
                min_y=min_y,
                max_y=max_y,
                node_min_x=split_x,
                node_max_x=node_max_x,
                node_min_y=split_y,
                node_max_y=node_max_y,
                trace=trace,
            )

class QuadTree(Generic[_V]):
    root: _QuadTreeNode[_V]
    max_leaf_size: int

    def __init__(
        self,
        v: Iterable[tuple[float, float, _V]] = (), *,
        max_leaf_size = 16,
    ):
        self.max_leaf_size = max_leaf_size
        if isinstance(v, _QuadTreeNode):
            self.root = v
        else:
            self.root = _QuadTreeNode.leaf(log2_size=0)
            for x, y, value in v:
                self.insert(x, y, value)

    def __repr__(self):
        return f"QuadTree({self.root})"

    def __iter__(self) -> Iterator[tuple[float, float, _V]]:
        return self.root.__iter__()

    def expand_root(self, target_log2_size: int, *, trace: None | Callable[[str]] = None):
        root = self.root
        if target_log2_size > root.log2_size and root.values is not None:
            # leaf node -- just expand in place
            if trace is not None:
                trace(f"expand_root({target_log2_size}) leaf")
            root.log2_size = target_log2_size
        while target_log2_size > root.log2_size:
            if trace is not None:
                trace(f"expand_root({target_log2_size}) interior: {root}")
            if root.nx_ny is not None:
                root.nx_ny = _QuadTreeNode.interior(
                    log2_size=root.log2_size,
                    px_py=root.nx_ny,
                )
            if root.px_ny is not None:
                root.px_ny = _QuadTreeNode.interior(
                    log2_size=root.log2_size,
                    nx_py=root.px_ny,
                )
            if root.nx_py is not None:
                root.nx_py = _QuadTreeNode.interior(
                    log2_size=root.log2_size,
                    px_ny=root.nx_py,
                )
            if root.px_py is not None:
                root.px_py = _QuadTreeNode.interior(
                    log2_size=root.log2_size,
                    nx_ny=root.px_py,
                )
            root.log2_size += 1
        if trace is not None:
            trace(f"expand_root({target_log2_size}) done: {root}")

    def range(
        self,
        min_x: float,
        max_x: float,
        min_y: float,
        max_y: float, *,
        trace: None | Callable[[str]] = None,
    ) -> Iterable[tuple[float, float, _V]]:
        assert not (isnan(min_x) or isnan(max_x)
                    or isnan(min_y) or isnan(max_y))
        if min_x > max_x or min_y > max_y:
            return ()
        size = ldexp(1.0, self.root.log2_size)
        return self.root.range(
            min_x=min_x,
            max_x=max_x,
            min_y=min_y,
            max_y=max_y,
            node_min_x=-size,
            node_max_x=size,
            node_min_y=-size,
            node_max_y=size,
            trace=trace,
        )

    def insert(self, x: float, y: float, value: _V, *, trace: None | Callable[[str]] = None):
        assert isfinite(x) and isfinite(y), "invalid coordinates"
        if trace is not None:
            trace(f"insert({x}, {y}, {value!r})")
        abs_x = abs(x)
        abs_y = abs(y)
        abs_max = max(abs_x, abs_y)
        _, log2_abs_max = frexp(abs_max)
        if trace is not None:
            trace(f"insert({x}, {y}, _): log2_abs_max={log2_abs_max}")
        self.expand_root(log2_abs_max, trace=trace)
        node = self.root
        split_x = 0.0
        split_y = 0.0
        delta = ldexp(0.5, node.log2_size)
        # walk down tree
        while node.values is None:
            if trace is not None:
                trace(f"insert({x}, {y}, _): interior: split_x={split_x} "
                    f"split_y={split_y} delta={delta} node={node}")
            if x < split_x:
                if y < split_y:
                    if node.nx_ny is None:
                        node.nx_ny = _QuadTreeNode.leaf(node.log2_size - 1, [(x, y, value)])
                        if trace is not None:
                            trace(f"insert({x}, {y}, _): insert nx_ny leaf: node={node}")
                        return
                    node = node.nx_ny
                    split_y -= delta
                else:
                    if node.nx_py is None:
                        node.nx_py = _QuadTreeNode.leaf(node.log2_size - 1, [(x, y, value)])
                        if trace is not None:
                            trace(f"insert({x}, {y}, _): insert nx_py leaf: node={node}")
                        return
                    node = node.nx_py
                    split_y += delta
                split_x -= delta
            else:
                if y < split_y:
                    if node.px_ny is None:
                        node.px_ny = _QuadTreeNode.leaf(node.log2_size - 1, [(x, y, value)])
                        if trace is not None:
                            trace(f"insert({x}, {y}, _): insert px_ny leaf: node={node}")
                        return
                    node = node.px_ny
                    split_y -= delta
                else:
                    if node.px_py is None:
                        node.px_py = _QuadTreeNode.leaf(node.log2_size - 1, [(x, y, value)])
                        if trace is not None:
                            trace(f"insert({x}, {y}, _): insert px_py leaf: node={node}")
                        return
                    node = node.px_py
                    split_y += delta
                split_x += delta
            delta *= 0.5
        # got to a leaf
        node.values.append((x, y, value))
        if trace is not None:
            trace(f"insert({x}, {y}, _): leaf node={node}")
        node.split(
            split_x=split_x,
            split_y=split_y,
            delta=delta,
            max_leaf_size=self.max_leaf_size,
            trace=trace,
        )

class _QuadTreeTest(unittest.TestCase):
    def test_quad_tree_range(self):
        items = []
        for x in range(8):
            x += 0.5
            x /= 4
            for y in range(8):
                y += 0.5
                y /= 4
                items.append((x, y, f"{x},{y}"))
        items.sort()
        qt = QuadTree(items)
        def check_range(min_x: float, max_x: float, min_y: float, max_y: float):
            with self.subTest(min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y):
                expected = list(filter(
                    lambda i: min_x <= i[0] <= max_x and min_y <= i[1] <= max_y, items))
                range_result = sorted(qt.range(
                    min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y, trace=print))
                print(f"expected={expected}")
                print(f"range_result={range_result}")
                self.assertEqual(expected, range_result)
        check_range(-1, 1, -1, 1)
        check_range(0.125, 0.25, 0.125, 0.25)
        check_range(items[0][0], items[0][0], items[0][1], items[0][1])
        check_range(1, 0, -1, 1)

    def test_quad_tree_insert(self):
        qt = QuadTree(max_leaf_size=4)
        for x in range(8):
            x += 0.5
            x /= 4
            for y in range(8):
                y += 0.5
                y /= 4
                qt.insert(x, y, f"{x},{y}")
        self.assertEqual(repr(qt), """QuadTree(interior(
    log2_size=1,
    px_py=interior(
        log2_size=0,
        nx_ny=interior(
            log2_size=-1,
            nx_ny=leaf(
                log2_size=-2,
                values=[
                    (0.125, 0.125, '0.125,0.125'),
                    (0.125, 0.375, '0.125,0.375'),
                    (0.375, 0.125, '0.375,0.125'),
                    (0.375, 0.375, '0.375,0.375'),
                ],
            ),
            px_ny=leaf(
                log2_size=-2,
                values=[
                    (0.625, 0.125, '0.625,0.125'),
                    (0.625, 0.375, '0.625,0.375'),
                    (0.875, 0.125, '0.875,0.125'),
                    (0.875, 0.375, '0.875,0.375'),
                ],
            ),
            nx_py=leaf(
                log2_size=-2,
                values=[
                    (0.125, 0.625, '0.125,0.625'),
                    (0.125, 0.875, '0.125,0.875'),
                    (0.375, 0.625, '0.375,0.625'),
                    (0.375, 0.875, '0.375,0.875'),
                ],
            ),
            px_py=leaf(
                log2_size=-2,
                values=[
                    (0.625, 0.625, '0.625,0.625'),
                    (0.625, 0.875, '0.625,0.875'),
                    (0.875, 0.625, '0.875,0.625'),
                    (0.875, 0.875, '0.875,0.875'),
                ],
            ),
        ),
        px_ny=interior(
            log2_size=-1,
            nx_ny=leaf(
                log2_size=-2,
                values=[
                    (1.125, 0.125, '1.125,0.125'),
                    (1.125, 0.375, '1.125,0.375'),
                    (1.375, 0.125, '1.375,0.125'),
                    (1.375, 0.375, '1.375,0.375'),
                ],
            ),
            px_ny=leaf(
                log2_size=-2,
                values=[
                    (1.625, 0.125, '1.625,0.125'),
                    (1.625, 0.375, '1.625,0.375'),
                    (1.875, 0.125, '1.875,0.125'),
                    (1.875, 0.375, '1.875,0.375'),
                ],
            ),
            nx_py=leaf(
                log2_size=-2,
                values=[
                    (1.125, 0.625, '1.125,0.625'),
                    (1.125, 0.875, '1.125,0.875'),
                    (1.375, 0.625, '1.375,0.625'),
                    (1.375, 0.875, '1.375,0.875'),
                ],
            ),
            px_py=leaf(
                log2_size=-2,
                values=[
                    (1.625, 0.625, '1.625,0.625'),
                    (1.625, 0.875, '1.625,0.875'),
                    (1.875, 0.625, '1.875,0.625'),
                    (1.875, 0.875, '1.875,0.875'),
                ],
            ),
        ),
        nx_py=interior(
            log2_size=-1,
            nx_ny=leaf(
                log2_size=-2,
                values=[
                    (0.125, 1.125, '0.125,1.125'),
                    (0.125, 1.375, '0.125,1.375'),
                    (0.375, 1.125, '0.375,1.125'),
                    (0.375, 1.375, '0.375,1.375'),
                ],
            ),
            px_ny=leaf(
                log2_size=-2,
                values=[
                    (0.625, 1.125, '0.625,1.125'),
                    (0.625, 1.375, '0.625,1.375'),
                    (0.875, 1.125, '0.875,1.125'),
                    (0.875, 1.375, '0.875,1.375'),
                ],
            ),
            nx_py=leaf(
                log2_size=-2,
                values=[
                    (0.125, 1.625, '0.125,1.625'),
                    (0.125, 1.875, '0.125,1.875'),
                    (0.375, 1.625, '0.375,1.625'),
                    (0.375, 1.875, '0.375,1.875'),
                ],
            ),
            px_py=leaf(
                log2_size=-2,
                values=[
                    (0.625, 1.625, '0.625,1.625'),
                    (0.625, 1.875, '0.625,1.875'),
                    (0.875, 1.625, '0.875,1.625'),
                    (0.875, 1.875, '0.875,1.875'),
                ],
            ),
        ),
        px_py=interior(
            log2_size=-1,
            nx_ny=leaf(
                log2_size=-2,
                values=[
                    (1.125, 1.125, '1.125,1.125'),
                    (1.125, 1.375, '1.125,1.375'),
                    (1.375, 1.125, '1.375,1.125'),
                    (1.375, 1.375, '1.375,1.375'),
                ],
            ),
            px_ny=leaf(
                log2_size=-2,
                values=[
                    (1.625, 1.125, '1.625,1.125'),
                    (1.625, 1.375, '1.625,1.375'),
                    (1.875, 1.125, '1.875,1.125'),
                    (1.875, 1.375, '1.875,1.375'),
                ],
            ),
            nx_py=leaf(
                log2_size=-2,
                values=[
                    (1.125, 1.625, '1.125,1.625'),
                    (1.125, 1.875, '1.125,1.875'),
                    (1.375, 1.625, '1.375,1.625'),
                    (1.375, 1.875, '1.375,1.875'),
                ],
            ),
            px_py=leaf(
                log2_size=-2,
                values=[
                    (1.625, 1.625, '1.625,1.625'),
                    (1.625, 1.875, '1.625,1.875'),
                    (1.875, 1.625, '1.875,1.625'),
                    (1.875, 1.875, '1.875,1.875'),
                ],
            ),
        ),
    ),
))""")