from collections import abc
from typing import Callable, Generic, Iterable, Iterator, Protocol, TypeAlias, TypeVar, overload


_T = TypeVar("_T")
_T_contra = TypeVar("_T_contra", contravariant=True)

class _SupportsLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra, /) -> bool: ...

class _SupportsGT(Protocol[_T_contra]):
    def __gt__(self, other: _T_contra, /) -> bool: ...

_SupportsRichComparison: TypeAlias = _SupportsLT | _SupportsGT

class SetById(abc.MutableSet[_T], Generic[_T]):
    __slots__ = "__data",

    def __init__(self, items: None | Iterable[_T]=None, /):
        if items is None:
            items = ()
        self.__data = {id(i): i for i in items}

    def __contains__(self, x: object) -> bool:
        return id(x) in self.__data

    def __iter__(self) -> Iterator[_T]:
        return iter(self.__data.values())

    def __len__(self) -> int:
        return len(self.__data)

    def add(self, value: _T) -> None:
        self.__data[id(value)] = value

    def discard(self, value: _T) -> None:
        key = id(value)
        data = self.__data
        if key in data:
            del data[key]

    def remove(self, value: _T) -> None:
        try:
            del self.__data[id(value)]
        except KeyError:
            raise KeyError(value) from None

    def pop(self) -> _T:
        return self.__data.popitem()[1]

    def clear(self) -> None:
        return self.__data.clear()

    def __repr__(self) -> str:
        if len(self.__data) == 0:
            return "SetById()"
        return f"SetById([{', '.join(map(repr, self))}])"

    @overload
    def sort(
        self, *,
        key: None = None,
        reverse: bool = False,
    ): ...
    @overload
    def sort(
        self, *,
        key: Callable[[_T], _SupportsRichComparison],
        reverse: bool = False,
    ): ...
    def sort(
        self, *,
        key: None | Callable[[_T], _SupportsRichComparison] = None,
        reverse: bool = False,
    ):
        adj_key = None if key is None else lambda kv: key(kv[1])
        self.__data = dict(sorted(self.__data.items(), key=adj_key, reverse=reverse))

